package Net::OICQ;

# $Id: OICQ.pm,v 1.26 2003/06/30 03:19:40 tans Exp $

# Copyright (c) 2002 Shufeng Tan.  All rights reserved.
# 
# This package is free software and is provided "as is" without express
# or implied warranty.  It may be used, redistributed and/or modified
# under the terms of the Perl Artistic License (see
# http://www.perl.com/perl/misc/Artistic.html)

use 5.006;
use strict;
use warnings;
use Carp;
use FileHandle;
use Digest::MD5;
use IO::Socket;

require Exporter;
require DynaLoader;

our @ISA = qw(Exporter DynaLoader);

our $VERSION = '0.7';

bootstrap Net::OICQ $VERSION;

#################### Begin OICQ protocol data ######################

# Byte 0x00 is always 0x02, i suspect it is protocol version

our $PROTO_VER = "\x02";

# Bytes 0x01-0x02 indicates packet source

# 0x01 0x00 for packets from server
# 0x06 0x2e for packets from GB client version 2000c build 630
# 0x07 0x2e for packets from En client version 2000c build 305
# 0x08 0x01 for packets from En client version 2000c build 630

our $SRC_SERVER = "\x01\x00";
our $SRC_CLIENT = "\x08\x01"; # English version 2000c build 630

# Byte 0x03 is always 0x00, maybe reserved for future commands?
# Byte 0x04 is command

our %CmdCode = (
    logout             => "\0\x01",
    keep_alive         => "\0\x02",
    reg_new_id_2       => "\0\x03",
    update_info        => "\0\x04",
    search_users       => "\0\x05",
    get_user_info      => "\0\x06",
    add_contact_1      => "\0\x09",
    del_contact        => "\0\x0a",
    add_contact_2      => "\0\x0b",
    set_mode           => "\0\x0d",
    reg_new_id_1       => "\0\x11",
    ack_service_msg    => "\0\x12",
    send_msg           => "\0\x16",
    recv_msg           => "\0\x17",
    forbid_contact     => "\0\x1c",
    login              => "\0\x22",
    get_friends_list   => "\0\x26",
    get_online_friends => "\0\x27",
    recv_service_msg   => "\0\x80",
    recv_friend_status => "\0\x81",
);

# Bytes 0x05-0x06 form a packet sequence number, a 16-bit integer

# Message precursor
our $MsgPrecursor     = "\0"x3 . "\x01" . "\0"x4 . "\x01";
our $AutoMsgPrecursor = "\0"x3 . "\x01" . "\0"x4 . "\x02";
our $MsgPrecursor_old = "\0"x8 . "\x01";

# Login modes
our %ConnectMode = (
    Normal    => "\x0a",
    Invisible => "\x28"
);

# System message code for 0x80 cmd
our %ServiceMsgCode = (
    '01' => 'User',
    '02' => 'ContactRequest',
    '06' => 'Broadcast'
);

# Separators
our $FS = "\x1e";   # Field separator
our $RS = "\x1f";   # Record separator

our @InfoHeader = qw(
    UserID Nickname Country Province PostCode Street Phone Age Sex Realname
    Email PagerCode PagerProvider PagerStationNum PagerNum PagerType
    Occupation Homepage Authorization unkn19 unkn20 Avatar
    MobilePhone MobileType Aboutme City unkn26 unkn27 unkn28 PublishMobile
    PublishContact School Horoscope Shengxiao BloodType unkn35 unkn36
);

# Some constants for constructing client packets
my $PacketHead = $PROTO_VER . $SRC_CLIENT;
my $PacketTail = "\x03";

my $C_LOGIN_17_33 = pack('H*', join("", qw(
00 00 00 09 00 00 00 01 00 00 00 11 f3 c8 9b c5
1f 24 ce d2 9e fe de ba cc 52 b4 b3 02
)));

my $C_LOGIN_35_END = pack('H*', join("", qw(
4d ea b4 a9 e6 85 b0 46 bb 0b 66 a0 1a f4 b6 b2
)));

#################### End OICQ protocol data ########################

my @ServerIPs = qw(
    61.144.238.146     202.104.129.251    202.104.129.254     61.141.194.203
    202.104.129.252    202.104.129.253     61.144.238.145     218.17.209.22
);

# Methods that do not require connection to a server

sub new {
    my ($class) = @_;
    my $dir = "$ENV{HOME}/.oicq";
    if (-e $dir) {
        -d $dir or croak "$dir exists but is not a directory";
    } else {
        mkdir($dir) or croak "Failed to mkdir $dir: $!";
    }
    my $self = {
        Dir        => $dir,
        LastSvrAck => 0,
        Debug      => 0  # 1 - trace packets, 2 - desect packets
    };
    my $logfile = "$dir/oicq.log";
    my $log = new FileHandle ">>$logfile";
    defined($log) or croak "Failed to open >>$logfile";
    $log->autoflush;
    $self->{LogFile} = $logfile;
    $self->{Log} = $log;
    return bless($self, $class);
}

sub set_user {
    my ($self, $id, $pw) = @_;

    $self->{Id} = $id;
    $self->{Passwd} = $pw;
    $self->{_Id} = pack('N', $id);
    $self->{PWMD5} = Digest::MD5::md5($pw);
    $self->{EventQueue} = [];
    $self->{SearchCount} = 0;
    $self->{LogChat}   = 1;
    $self->{Info}      = {};
    $self->{Away}      = 0;
    $self->{AutoReply} = {};  # use id as hash key
    $self->{AutoAwayTime} = "";

    my $userdir = "$self->{Dir}/$id";
    -e $userdir or mkdir($userdir);
    if (-d $userdir) {
        foreach ($self->get_saved_ids) { $self->get_nickname($_) };
        my $logfile = "$userdir/user.log";
        my $log = new FileHandle(">>$logfile");
        if (defined $log) {
            $self->log_t("Switch log to $logfile") if $self->{Debug};
            $self->{Log} = undef;
            $self->{LogFile} = $logfile;
            $self->{Log} = $log;
            $log->autoflush;
        } else {
            $self->log_t("Failed to open >>$logfile");
        }
    } else {
        $self->log_t("Failed to mkdir $userdir");
    }
}

sub decrypt {
    my ($self, $crypt, $key) = @_;

    my $crypt_len = length($crypt);
    if (($crypt_len % 8) || ($crypt_len < 16)) {
        $self->log("Invalid input to decrypt: len = $crypt_len\n");
        return undef;
    }

    my $c_buf = substr($crypt, 0, 8);
    my $p_buf = "";
    Net::OICQ::crypt($c_buf, $key, 1, $p_buf);
    my $pad_len = ord(substr($p_buf, 0, 1) & "\007");
    my $plain_len = $crypt_len - $pad_len - 10;
    my $plain = $p_buf;
    my $pre_plain = $p_buf;
    my $pre_crypt = $c_buf;

    for (my $i = 8; $i < $crypt_len; $i += 8) {
        $c_buf = substr($crypt, $i, 8);
        Net::OICQ::crypt($c_buf ^ $pre_plain, $key, 1, $p_buf);
        $pre_plain = $p_buf;
        $p_buf ^= $pre_crypt;
        $plain .= $p_buf;
        $pre_crypt = $c_buf;
    }
    if (substr($plain, -7, 7) ne "\0\0\0\0\0\0\0") {
        return undef unless $self->{Debug};
        $self->log("Decrypt error, crypt and plain data dumped\n",
                   $self->hexdump($crypt), "\n", $self->hexdump($plain),
                   '#'x70, "\n");
        return undef;
    }
    return substr($plain, -7-$plain_len, $plain_len);
}

# Methods for building OICQ packets

sub encrypt {
    my ($self, $plain, $key) = @_;
    my $plain_len = length($plain);
    my $head_pad_len = ($plain_len + 10) % 8;
    $head_pad_len = 8 - $head_pad_len if $head_pad_len;
    my $padded_plain = chr(0xa8 + $head_pad_len) .
       #rand_str(2+$head_pad_len) .
       (chr(0xad) x (2 + $head_pad_len)) .
                       $plain . ("\0" x 7);
    my $padded_plain_len = length($padded_plain);
    my $crypt = "";
    my $pre_plain = "\0" x 8;
    my $pre_crypt = $pre_plain;
    for (my $i = 0; $i < $padded_plain_len; $i += 8) {
        my $p_buf = substr($padded_plain, $i, 8) ^ $pre_crypt;
	my $c_buf = "";
        Net::OICQ::crypt($p_buf, $key, 0, $c_buf);
	$c_buf ^= $pre_plain;
	$crypt .= $c_buf;
	$pre_crypt = $c_buf;
	$pre_plain = $p_buf;
    }
    return $crypt;
}

sub build_packet {
    my ($self, $cmd, $data) = @_;
    croak "No such command: $cmd" unless exists $CmdCode{$cmd};
    $self->{Seq}++;
    $PacketHead . $CmdCode{$cmd} . pack('n', $self->{Seq}) . $self->{_Id} .
    $self->encrypt($data, $self->{Key}) . $PacketTail;
}

sub rand_str {
    my $len = pop;
    join('', map(pack("C", rand(0xff)), 1..$len));
}

sub build_login_packet {
    my ($self) = @_;
    my $seq = "\0" . rand_str(1);
    $self->{Seq} = unpack('n', $seq);

    my $randkey = rand_str(16);
    $self->{RandKey} = $randkey;

    # 0x00-0x10
    my $data = $self->encrypt("", $self->{PWMD5}) . "\0";
    # 0x11-0x22    We use fake ip 10.0.0.1
    $data .= pack('Nn', 0x0a000001, 4000);
    # 0x23-0x43
    $data .= $C_LOGIN_17_33 . $ConnectMode{$self->{ConnectMode}} .
             $C_LOGIN_35_END;

    $PacketHead . $CmdCode{'login'} . $seq . $self->{_Id} . $randkey .
    $self->encrypt($data, $randkey) . $PacketTail;
}

sub build_logout_packet {
    my ($self) = @_;
    $PacketHead . $CmdCode{'logout'} . ("\xff" x 2) . $self->{_Id} .
    $self->encrypt($self->{PWMD5}, $self->{Key}) . $PacketTail;
}

# Tencent no longer allows getting new number this way
# So these two are pretty much useless

sub build_reg_1_packet {
    my ($self) = @_;
    my $randkey = rand_str(16);
    my $data = "";   # fixme
    $PacketHead . $CmdCode{'reg_new_id_1'} . pack('n', 1) . "\0"x4 .
    $randkey . $self->encrypt($data, $randkey) . $PacketTail;
}

sub build_reg_2_packet {
    my ($self) = @_;
    my $randkey = rand_str(16);
    my $data = "";   # fixme
    $PacketHead . $CmdCode{'reg_new_id_2'} . pack('n', 2) . "\0"x4 .
    $randkey . $self->encrypt($data, $randkey) . $PacketTail;
}

# Methods for logging and output

sub log {
    my $self = shift;
    my $log = $self->{Log};
    print $log @_;
}

sub logf {
    my $self = shift;
    my $log = $self->{Log};
    printf $log @_;
}

sub log_t {
    my ($self, @msg) = @_;
    my $log = $self->{Log};
    print $log substr(localtime, 4, 16), @msg, "\n";
}

sub hexdump {
    my $str = pop;
    return unless defined $str;
    my $res = "";
    my $len = length($str);
    for (my $i = 0; $i < $len; $i += 16) {
        my $s = substr($str, $i, 16);
        my $hex = unpack('H*', $s);
        $s =~ s/[\x00-\x1f]/./g;   # 0x00-0x1f will screw up terminal
        $hex =~ s/(\w\w)/$1 /g;
        $res .= sprintf("%-48s    %s\n", $hex, $s);
    }
    return $res;
}

sub dump_substr {
    my ($self, $data, $tmpl, $prefix, $begin, $len) = @_;
    my ($str, $end);
    if (defined($len)) {
        $str = substr($data, $begin, $len);
        $end = ($begin+$len < length($data)) ? $begin+$len-1 : length($data)-1;
    } else {
        $str = substr($data, $begin);
        $end = length($data)-1;
    }
    $self->logf("0x%02x-0x%02x %s: ", $begin, $end, $prefix);
    if ($tmpl =~ /\w/) {
        if ($tmpl eq 'H*') {
            $self->log("\n", $self->hexdump($str));
        } else {
            $self->log(unpack($tmpl, $str), "\n");
        }
    } else {
        $self->log("$str\n");
    }
}

sub desect {
    my $self = shift;
    return unless $self->{Debug} > 1;
    my $data = shift;
    foreach my $arg (@_) {
        $self->dump_substr($data, @{$arg});
    }
    return;
}

sub remove_saved_id {
    my ($self, $id) = @_;
    my $file = "$self->{Dir}/$self->{Id}/$id.dat";
    if (-e $file) {
        unlink($file);
        return 0 if -e $file;
        return 1;
    } else {
        return 0;
    }
}

sub get_saved_ids {
    my ($self) = @_;
    my $dir = "$self->{Dir}/$self->{Id}";
    my @ids = ();
    if (opendir(DIR, $dir)) {
        while(my $f = readdir(DIR)) {
            next unless $f =~ /^(\d+)\.dat$/;
            push @ids, $1;
        }
        closedir(DIR);
    }
    return @ids;
}

sub get_face {
    my $num = pop;
    sprintf('%d-%d', 1 + $num/3, 1 + $num % 3);
}

sub toggle_autoreply {
    my ($self) = @_;
    if ($self->{Away}) {
        $self->{Away} = 0;
        return "off";
    } else {
        $self->{Away} = 1;
        return "on";
    }
}

# Nickname can be updated by get_friends_list or get_user_info

sub get_nickname {
    my ($self, $id) = @_;
    if (defined $self->{Info}->{$id}) {
        if (defined $self->{Info}->{$id}->{Nickname}) {
            return $self->{Info}->{$id}->{Nickname};
        }
    } else {
        $self->{Info}->{$id} = {};
    }
    my $infofile = "$self->{Dir}/$self->{Id}/$id.dat";
    my $nick = "";
    if (open(INFO, $infofile)) {
        while(my $line = <INFO>) {
            if ($line =~ /^Nickname +=> *'(.*)'/) {
                $nick = $1;
                last;
            }
        }
        close(INFO);        
    }
    $self->{Info}->{$id}->{Nickname} = $nick;
    return $nick;
}

sub get_servers {
    my @servers;
    if (exists $ENV{OICQ_SVR} and $ENV{OICQ_SVR} =~ /\d+/) {
        my $svr = $ENV{OICQ_SVR};
        $svr =~ s/^\D+//;
        $svr =~ s/\D+$//;
        @servers = split(/[^\w\.]+/, $svr);
        return @servers if @servers;
    }
        
    foreach my $n (2..7, "") {
        my $ip = gethostbyname("sz$n.tencent.com");
        next unless $ip;
        push @servers, join('.', map(ord, split('', $ip)));
    }
    if (@servers) {
        # this unregistered ip seems to be for mobile numbers only
        push @servers, '218.17.209.22';
        return @servers;
    }
    return @ServerIPs;
}

sub connect {
    my ($self, $ip, $port) = @_;
    croak "Server IP not provided"  unless defined($ip) && $ip;
    $port = 8000 unless defined($port) && $port;

    my $socket = IO::Socket::INET->new(
            Proto => 'udp', PeerAddr => $ip, PeerPort => $port
    ) or croak("socket error: $@");

    $self->{SvrIP}   = $ip;
    $self->{SvrPort} = $port;
    $self->{Socket}  = $socket;
    return $socket;
}

sub send2svr {
    my ($self, $cmd, $data) = @_;
    my $packet = $self->build_packet($cmd, $data);
    Net::OICQ::ClientEvent->new($packet, $self)->process;
    $self->{Socket}->send($packet);
}

# get_friends_list provided by Chen Peng

sub get_friends_list {
    my ($self, $flag) = @_;
    defined $flag or $flag = "\0\0";
    $self->send2svr('get_friends_list', $flag);
}

sub get_online_friends {
    my ($self) = @_;
    $self->send2svr('get_online_friends', "\x02\0\0\0\0");
}

sub set_mode {
    my ($self, $mode_code) = @_;
    $self->send2svr('set_mode', $mode_code);
}

sub get_user_info {
    my ($self, $id) = @_;
    $self->send2svr('get_user_info', $id);
}

sub update_info {
    my ($self, $hashref) = @_;
    my $info = $self->{MyInfo};
    return unless defined $hashref and defined $info;
    my %new_info;
    # Use all upper-case letters for keys
    foreach my $k (keys %$hashref) {
	$new_info{uc($k)} = $hashref->{$k};
    }
    my @update;
    for (my $i = 1; $i < $#InfoHeader; $i++) {
        my $attr = uc($InfoHeader[$i]);
        push(@update, defined($new_info{$attr}) ? $new_info{$attr} : $info->[$i]);
    }
    $self->send2svr('update_info', join($RS, "", "", @update));
    return 1;
}

sub set_passwd {
    my ($self, $newpw) = @_;
    return unless defined $self->{MyInfo};
    my @info = @{$self->{MyInfo}};
    pop @info; shift @info;
    $self->send2svr('update_info', join($RS, $self->{Passwd}, $newpw, @info));
    return 1;
}

sub accept_contact {
    my ($self, $id) = @_;
    $self->send2svr('add_contact_2', $id.$RS."0");
}

sub reject_contact {
    my ($self, $id) = @_;
    $self->send2svr('add_contact_2', $id.$RS."1");
}

sub add_contact {
    my ($self, $id) = @_;
    $self->send2svr('add_contact_1', "$id");
    $self->{ContactId} = $id;
}

sub add_contact_2 {
    my ($self, $msg) = @_;
    my $mmid = $self->{ContactId};
    return unless defined $mmid;
    $self->send2svr('add_contact_2', "$mmid".$RS."2".$RS."$msg");
    delete $self->{ContactId};
}

sub del_contact {
    my ($self, $id) = @_;
    $self->send2svr('del_contact', "$id");
}

sub forbid_contact {
    my ($self, $id) = @_;
    $self->send2svr('forbid_contact', "$id");
}

sub send_msg {
    my ($self, $dstid, $msg) = @_;
    $self->log_t("Sent message to $dstid:\n", $msg) if $self->{LogChat} or $self->{Debug};
    $dstid = pack('N', $dstid);
    my $data = $self->{_Id} .
               $dstid .
               $SRC_CLIENT .
               $self->{_Id} .
               $dstid .
               Digest::MD5::md5($dstid . $self->{Key}) .
               "\0\x0b" .
               pack('S', $self->{Seq}) .
               "\x3d" .
               rand_str(3) .
               "\0\0" . $MsgPrecursor .
               $msg;
    $self->send2svr('send_msg', $data);
}

sub ack_msg {
    my ($self, $plain) = @_;
    foreach (1..3) {
        $self->send2svr('recv_msg', substr($plain, 0, 12));
    }
}

sub ack_service_msg {
    my ($self, $code, $srcid, $seq) = @_;
    $self->send2svr('ack_service_msg', "$code$FS$srcid$FS$seq");
}

sub keepalive {
    my ($self) = @_;
    $self->send2svr('keep_alive', $self->{Id});
    $self->{LastKeepaliveTime} = time;
}

sub search_user {
    my ($self, $id) = @_;
    $self->send2svr('search_users', join($RS, '0', $id, '-','-','0'));
}

sub list_online_users {
    my ($self, $num) = @_;
    defined $num or $num = 1;
    my $begin = $self->{SearchCount};
    $self->{SearchCount} += $num; 
    my $end = $self->{SearchCount} -1;
    foreach my $p ($begin .. $end) {
        $self->send2svr('search_users', "1".$RS."$p");
    }
}

sub logout {
    my $self = shift;
    defined($self->{LogoutPacket}) && $self->{LogoutPacket} || return;
    my $packet = $self->{LogoutPacket};
    foreach (1..3) {
        $self->{Socket}->send($packet);
    }
}

sub process_packet {
    my ($self, $packet) = @_;

    my $ver = substr($packet, 0, 1);
    if ($ver ne $PROTO_VER) {
        $self->log_t('Unknown protocol version, ', unpack('H*', $ver));
        $self->log($self->hexdump($packet),"\n") if $self->{Debug};
        return;
    }

    my $src = substr($packet, 1, 2);
    my $event;
    if ($src eq $SRC_SERVER) {
        $event = Net::OICQ::ServerEvent->new($packet, $self);
        $self->{LastSvrAck} = time;
    } else {
        $event = Net::OICQ::ClientEvent->new($packet, $self);
    }

    return unless defined $event;

    if ($self->{Debug}) {
        $self->log_t('Seq=', unpack('H*', $event->{Seq}),
                     ', Src=', unpack('H*', $src),
                     ', Cmd=', unpack('H*', $event->{CmdCode}),
                     " ($event->{Cmd})");
    }

    return $event->process;
}

# Extract message from decrypted data using MsgPrecursor
sub get_msg {
    my ($self, $plain) = @_;
    if ($plain =~ /$MsgPrecursor(.*)$/s) {
        return $1;
    } elsif ($plain =~ /$MsgPrecursor_old(.*)$/s) {
        # Fucking tencent keeps changing
        return $1;
    } elsif ($plain =~ /$AutoMsgPrecursor(.*)$/s) {
        return "Auto-Reply:\n" . $1;
    } else {
        $self->log("Unable to get message from data:\n");
        $self->dumphex($plain);
        return undef;
    }
}

1;

__END__

=head1 NAME

Net::OICQ - Perl interface to an OICQ server

=head1 SYNOPSIS

  use Net::OICQ;
  $oicq = new Net::OICQ;

=head1 DESCRIPTION

=head2 EXPORT

None by default.

=head1 AUTHOR

Shufeng Tan <lt>perloicq@yahoo.com<gt>

=head1 SEE ALSO

L<perl>.

=cut
